# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json

import matplotlib.pyplot as plt
import numpy as np
from bert_schemas import job as job_schema
from bert_schemas import projected
from bert_schemas.projected import (
    BarrierHeights,
    BarrierPositions,
    BarrierTimes,
    BarrierWidths,
    ProjectedEnergies,
    ProjectedPositions,
)
from bert_schemas.projected import Settings as ProjectionSettings
from bert_schemas.timing import TimeMs, decimal_to_float

pset = ProjectionSettings()

# Module containing methods and classes for abstractions of quasi-1D "painted" light Oqtant object

# "ideal" optical potentials are those specified by the user using included objects
# "actual" optical potentials include implementation and hardware realities
# such as objects being projected as a sum of gaussians on a pre-defined position grid
# with dynamic weights recalculated on a periodic basis


class Snapshot(job_schema.Landscape):
    """A class that represents a painted optical landscape/potential at a single
    point in (manipulation stage) time
    """

    @classmethod
    def new(
        cls,
        time: TimeMs = 0.0,
        positions: ProjectedPositions = [-10, 10],
        potentials: ProjectedEnergies = [0, 0],
        interpolation: job_schema.InterpolationType = "LINEAR",
    ) -> Snapshot:
        """Method to create a new Snapshot object

        Args:
            time (float, optional): Time associated with the snapshot
            positions (list, optional): Position list for the snapshot
            potentials (list, optional): Potential energies corresponding to the list of positions
            interpolation (bert_schemas.job.InterpolationType, optional): How to connect the object's
                (positions, potentials) data in space.

        Returns:
            Snapshot: a new Snapshot object
        """
        return cls(
            time_ms=time,
            positions_um=positions,
            potentials_khz=potentials,
            spatial_interpolation=interpolation,
        )

    @classmethod
    def from_input(cls, landscape: job_schema.Landscape) -> Snapshot:
        """Method to create a Snapshot object from an existing jobs input

        Args:
            landscape (bert_schemas.job.Landscape): The input values

        Returns:
            Snapshot: A new Snapshot object created using the input data
        """
        return cls(**landscape.model_dump())

    def show_potential(
        self,
        xlimits: list[float] = [pset.min_position - 1, pset.max_position + 1],
        ylimits: list[float] = [-1.0, pset.max_energy + 1],
        include_ideal: bool = False,
    ) -> None:
        """Method to plot the potential energy as a function of position for a Landscape
        object at the given times

        Args:
            xlimits (list[float], optional): Plot limits for x axis
            ylimits (list[float], optional): Plot limits for y axis
            include_ideal (bool, optional): Flag for including target potential in plot
        """

        positions = np.arange(pset.min_position, pset.max_position + 0.1, 0.1, dtype=float)

        _, ax = plt.subplots()
        color = next(ax._get_lines.prop_cycler)["color"]
        (ln,) = plt.plot(positions, self.get_potential(positions=positions), color=color)
        if include_ideal:
            (ln2,) = plt.plot(
                positions,
                self.get_ideal_potential(positions=positions),
                "--",
                color=color,
            )
            plt.plot(self.positions_um, self.potentials_khz, ".", color=color)
            lns = [ln, ln2]
            labs = ["actual", "ideal"]
            ax.legend(lns, labs, loc=0)
        plt.xlabel("position (microns)", labelpad=6)
        plt.ylabel("potential energy (kHz)", labelpad=6)
        plt.title("Snapshot potential energy profile")
        plt.xlim(xlimits)
        plt.ylim(ylimits)
        plt.show()


# (potentially) dynamic landscape made up of snapshots
class Landscape(job_schema.OpticalLandscape):
    """Class that represents a dynamic painted-potential optical landscape constructed
    from individual (instantaneous time) Snapshots
    """

    @classmethod
    def new(
        cls,
        snapshots: list[Snapshot] = [
            Snapshot.new(time=0),
            Snapshot.new(time=2),
        ],
    ) -> Landscape:
        """Method to create a new Landscape object

        Args:
            snapshots (list[Snapshot], optional): A list of Snapshot objects

        Returns:
            Landscape: A new Landscape object
        """
        optical_landscapes = []
        for snapshot in snapshots:
            optical_landscapes.append(
                job_schema.Landscape(
                    time_ms=snapshot.time_ms,
                    positions_um=snapshot.positions_um,
                    potentials_khz=snapshot.potentials_khz,
                    spatial_interpolation=snapshot.spatial_interpolation,
                )
            )
        return cls(landscapes=optical_landscapes)

    @classmethod
    def from_input(cls, landscape: job_schema.OpticalLandscape) -> Landscape:
        """Method to create a Landscape object from an existing jobs input

        Args:
            landscape (job_schema.OpticalLandscape): The input values

        Returns:
            Landscape: A new Landscape object
        """
        return cls(**json.loads(landscape.model_dump_json()))

    # extract Snapshot abstract objects from backend data structure
    @property
    def snapshots(self) -> list[Snapshot]:
        """Property to get a list of Snapshot objects associated to a Landscape object

        Returns:
            list[Snapshot]: List of Snapshot objects
        """
        return [Snapshot(**landscape.model_dump()) for landscape in self.landscapes]

    def show_potential(
        self,
        times: list = [0.0],
        xlimits: list = [pset.min_position - 1, pset.max_position + 1],
        ylimits: list = [-1.0, pset.max_energy + 1],
        include_ideal: bool = False,
    ):
        """Method to plot the potential energy as a function of position for a Landscape object at the given times

        Args:
            times (list[float], optional): Times, in ms, at which to evaluate and plot the potential
            xlimits (list[float], optional): Plot limits for x axis
            ylimits (list[float], optional): Plot limits for y axis
            include_ideal (bool, optional): Flag for including target potential in plot
        """
        positions = np.arange(pset.min_position, pset.max_position + 0.1, 0.1, dtype=float)

        fig, ax = plt.subplots()
        lns = []
        labs = []
        for time in times:
            potentials = self.get_potential(time, positions)
            color = next(ax._get_lines.prop_cycler)["color"]
            (ln,) = plt.plot(positions, potentials, color=color)
            lns.append(ln)
            labs.append("t = " + str(time) + " ms")
            if include_ideal:
                potentials_ideal = self.get_ideal_potential(time=time, positions=positions)
                (ln,) = plt.plot(positions, potentials_ideal, "--", color=color)
                lns.append(ln)
                labs.append("t = " + str(time) + " ms (ideal)")

        plt.xlabel("position (microns)", labelpad=6)
        plt.ylabel("potential energy (kHz)", labelpad=6)
        plt.title("Landscape potential energy profile")
        plt.xlim(xlimits)
        plt.ylim(ylimits)
        ax.legend(lns, labs, loc=0)
        plt.show()


class Barrier(job_schema.Barrier):
    """Class that represents a painted optical barrier."""

    @classmethod
    def new(
        cls,
        positions: BarrierPositions = [0.0, 0.0],
        heights: BarrierHeights = [0.0, 0.0],
        widths: BarrierWidths = [1.0, 1.0],
        times: BarrierTimes = [0.0, 10.0],
        shape: job_schema.ShapeType = job_schema.ShapeType.GAUSSIAN,
        interpolation: job_schema.InterpolationType = job_schema.InterpolationType.LINEAR,
    ) -> Barrier:
        """Method to create a new Barrier object

        Args:
            positions (list[float], optional): Positions for the barrier
            heights (list[float], optional): Heights for the barrier
            widths (list[float], optional): Widths for the barrier
            times (list[float], optional): Times for the barrier
            shape (bert_schemas.job.ShapeType, optional): Shape of the barrier
            interpolation (bert_schemas.job.InterpolationType, optional): Interpolation type of the barrier

        Returns:
            Barrier: A new Barrier object

        Raises:
            ValueError: if data lists are not of equal length
        """
        if not (len(positions) == len(heights) == len(widths) == len(times)):
            raise ValueError("Barrier data lists must be of equal length, default minimum of 2")

        data = {
            "times_ms": times,
            "positions_um": positions,
            "heights_khz": heights,
            "widths_um": widths,
            "shape": shape,
            "interpolation": interpolation,
        }

        return cls(**data)

    @classmethod
    def from_input(cls, barrier: job_schema.Barrier) -> Barrier:
        """Method to create a Barrier object using the input values of a job

        Args:
            barrier (job_schema.Barrier): The input values

        Returns:
            Barrier: A new Barrier object created using the input data
        """
        return cls(**barrier.model_dump())

    def show_dynamics(self) -> None:
        """Method to plot the position, width and height of a Barrier object over time"""
        tstart = min(self.times_ms)
        tstop = max(self.times_ms)
        times = np.linspace(
            decimal_to_float(tstart),
            decimal_to_float(tstop),
            num = int((tstop - tstart) / pset.update_period),
            endpoint=True,
        )
        fig, ax1 = plt.subplots()

        # plot position and width vs time
        style = "steps-pre"
        color = next(ax1._get_lines.prop_cycler)["color"]
        ax1.set_xlabel("time (ms)")
        ax1.set_ylabel("position or width (microns)")
        ax1.set_xlim([-1, self.times_ms[-1] + 1])
        ax1.set_ylim([pset.min_position - 1, pset.max_position + 1])
        (ln1,) = plt.plot(times, self.get_positions(times), color=color, drawstyle=style)
        plt.plot(
            self.times_ms,
            self.get_positions(self.times_ms),
            ".",
            color=color,
        )
        color = next(ax1._get_lines.prop_cycler)["color"]
        (ln2,) = plt.plot(times, self.get_widths(times), color=color, drawstyle=style)
        plt.plot(self.times_ms, self.get_widths(self.times_ms), ".", color=color)

        # plot height on the same time axis
        ax2 = ax1.twinx()
        ax2.set_ylabel("height (kHz)")
        ax2.set_ylim([0, 100])
        color = next(ax1._get_lines.prop_cycler)["color"]
        (ln3,) = plt.plot(times, self.get_heights(times), color=color, drawstyle=style)
        plt.plot(self.times_ms, self.get_heights(self.times_ms), ".", color=color)

        # shared setup
        color = next(ax1._get_lines.prop_cycler)["color"]
        ax1.legend([ln1, ln2, ln3], ["position", "width", "height"], loc="upper left")
        plt.title("Barrier dynamics")
        fig.tight_layout()
        plt.show()

    def show_potential(
        self,
        times: list[float] = [0.0],
        xlimits: list[float] = [pset.min_position - 1, pset.max_position + 1],
        ylimits: list[float] = [-1.0, pset.max_energy + 1],
        include_ideal: bool = False,
    ) -> None:
        """Method to plot the potential energy as a function of position for a Barrier object

        Args:
            times (list[float], optional): The times, in ms, at which the potential is evaluated
            xlimits (list[float], optional): Plot limits for x axis
            ylimits (list[float], optional): Plot limits for y axis
            include_ideal (bool, optional): Flag for including target potential in plot
        """

        positions = np.arange(np.floor(min(xlimits)), np.ceil(max(xlimits)) + 1, 0.1)

        fig, ax1 = plt.subplots()
        ax = plt.gca()
        lns = []
        labs = []
        for time in times:
            color = next(ax._get_lines.prop_cycler)["color"]
            potential = self.get_potential(time=time, positions=positions)
            (ln,) = plt.plot(positions, potential, color=color)
            lns.append(ln)
            labs.append("t = " + str(time) + " ms")
            if include_ideal:
                potentials_ideal = self.get_ideal_potential(time=time, positions=positions)
                (ln,) = plt.plot(positions, potentials_ideal, "--", color=color)
                lns.append(ln)
                labs.append("t = " + str(time) + "ms (ideal)")

        plt.xlabel("position (microns)", labelpad=6)
        plt.ylabel("potential energy (kHz)", labelpad=6)
        plt.xlim(xlimits)
        plt.ylim(ylimits)
        ax1.legend(lns, labs, loc=0)
        plt.show()


class Pulse(job_schema.Pulse):
    """Class that represents a terminator laser pulse"""

    @classmethod
    def new(
        cls,
        times_ms: list[TimeMs],
        intensities_mw_per_cm2: list[float] = [1, 1],
        detuning_mhz: float = 0,
        interpolation: job_schema.InterpolationType = job_schema.InterpolationType.OFF,
    ) -> Pulse:
        """Method to create a new terminator laser pulse

        Args:
            times_ms (list) : [turn on time, turn off time]
            intensities_mw_per_cm2 (list[float]) : intensities in mw per cm^2, default = [1,1]
            detuning_mhz (float) : laser frequency detuning from resonance, default = 0
            interpolation (job_schema.InterpolationType) : interpolation in time for intensity,
                default = job_schema.InterpolationType.OFF
        Returns:
            Pulse: A new Pulse object

        Raises:
            ValueError: #TODO check that no values are specified for fields which are not supported yet
        """

        print(times_ms, intensities_mw_per_cm2)

        data = {
            "times_ms": times_ms,
            "intensities_mw_per_cm2": intensities_mw_per_cm2,
            "detuning_mhz": detuning_mhz,
            "interpolation": interpolation,
        }

        return cls(**data)

    @classmethod
    def from_input(cls, pulse: job_schema.Pulse) -> Pulse:
        """Method to create a Pulse object using the input values of a job

        Args:
            pulse (job_schema.Pulse): The input values

        Returns:
            Pulse: A new Pulse object created using the input data
        """
        return cls(**pulse.model_dump())


class Laser(job_schema.Laser):
    """Class that represents a terminator laser with a single pulse."""

    @classmethod
    def new(
        cls,
        pulses: list[Pulse],
        type: job_schema.LaserType = "TERMINATOR",
        position_um: float = pset.terminator_position,
    ) -> Laser:
        """Method to create a new Laser

        Args:
            pulses (list[Pulse]) : a list of laser pulse objects
            type (job_schema.LaseType) : laser type by task/experiment, default = "TERMINATOR"
            position_um (float) : position along the X axis in microns, default = 0 #TODO put real beam center here

        Returns:
            Barrier: A new Barrier object

        Raises:
            OqtantError: #TODO check that there is only one pulse
        """

        data = {"type": type, "position_um": position_um, "pulses": pulses}

        return cls(**data)

    @classmethod
    def from_input(cls, laser: job_schema.Laser) -> Laser:
        """Method to create a Laser object using the input values of a job

        Args:
            laser (job_schema.Laser): The input values

        Returns:
            Laser: A new Laser object created using the input data
        """
        return cls(**laser.model_dump())

    def is_on(self, time_ms: float) -> bool:
        for pulse in self.pulses:
            if (time_ms >= decimal_to_float(pulse.times_ms[0])) & (
                time_ms <= decimal_to_float(pulse.times_ms[-1])
            ):
                return True
        return False



import os
import re
import subprocess

import click


def get_running_ports():
    """
    Get running ports of Jupyter servers

    Args:
        None
    Returns:
        list: list of running ports
    """
    res = subprocess.getoutput("jupyter notebook list")

    pattern = r"http://localhost:(\d+)"
    regex = re.compile(pattern)

    running_ports = regex.findall(res)
    return running_ports


@click.command()
@click.option("--port", default=8888)
def start_notebook(port):
    """
    Start Jupyter server, and kill previously running Jupyter servers if any exist

    Args:
        port (int, optional): Port to run the Jupyter server on
    Returns:
        None
    """
    running_ports = get_running_ports()
    if running_ports:
        resp = click.confirm(
            "There are running Jupyter servers on ports: "
            + ", ".join(str(port) for port in running_ports)
            + ". Do you want to stop them? (You will not be able to start and log into another instance of Qqtant without doing so.)"
        )
        if resp:
            os.system(f"killport {' '.join(running_ports)}")
            click.echo("Killed running Jupyter servers")

    subprocess.run(["jupyter", "notebook", "--port", str(port)])


if __name__ == "__main__":
    start_notebook()



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from math import floor, log10

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as opt
from bert_schemas import job as job_schema
from matplotlib.ticker import FormatStrFormatter, LinearLocator
from pydantic import BaseModel

from oqtant.schemas.job import print_keys
from oqtant.util.exceptions import (
    JobPlotFitError,
    JobPlotFitMismatchError,
    JobReadError,
)


class OutputImageType(str, Enum):
    TIME_OF_FLIGHT = "TIME_OF_FLIGHT"
    IN_TRAP = "IN_TRAP"
    MOT = "MOT"
    TIME_OF_FLIGHT_FIT = "TIME_OF_FLIGHT_FIT"


class AxisType(str, Enum):
    x = "x"
    y = "y"


def in_trap_check(func):
    def wrapper(*args, **kwargs):
        if args[0].image_type == OutputImageType.IN_TRAP:
            print("** not available for In-Trap output **")
        else:
            return func(*args, **kwargs)

    return wrapper


class OqtantOutput(BaseModel):
    """A class that represents the output of a job submitted to Oqtant"""

    @property
    def image_type(self):
        if hasattr(self, "it_plot"):
            return OutputImageType.IN_TRAP
        else:
            return OutputImageType.TIME_OF_FLIGHT

    @property
    def fields(self):
        """Method to print out the output fields for an OqtantOutput"""
        return print_keys(self.model_dump())

    @property
    def atom_statistics(self):
        """Property that prints out the atom statistics of a TIME_OF_FLIGHT image job's output"""
        if not self.image_type.IN_TRAP:
            print(f"Temperature (nK): {self.temperature_nk}")
            print(f"Total atoms : {self.tof_atom_number}")
            print(f"Condensed atoms : {self.condensed_atom_number}")
            print(f"Thermal atoms : {self.thermal_atom_number}")
        else:
            print(
                "** Atom statistics only available for TIME_OF_FLIGHT imaging in a BEC job. **"
            )

    @property
    def TOF(self) -> np.ndarray | None:
        """Property that returns the shaped time of flight (TOF) image of a job's output if it exists

        Returns:
            numpy.ndarray: The reshaped pixels of the TOF image
        """
        reshaped_pixels = None
        try:
            reshaped_pixels = np.array(self.tof_image.pixels).reshape(
                self.tof_image.rows, self.tof_image.columns
            )
        except Exception:
            print("** not an TOF image **")
        return reshaped_pixels

    @property
    def IT(self) -> np.ndarray | None:
        """Property that returns the shaped in-trap (IT) image of a job's output if it exists

        Returns:
            numpy.ndarray: The reshaped pixels of the IT image
        """
        reshaped_pixels = None
        try:
            reshaped_pixels = np.array(self.it_plot.pixels).reshape(
                self.it_plot.rows, self.it_plot.columns
            )
        except Exception:
            print("** not an IT image **")
        return reshaped_pixels

    @property
    @in_trap_check
    def temperature(self):
        """Property that returns the atom temperature of a job's output

        Returns:
            int: The atom temperature of a job's output
        """
        return self.temperature_nk

    @property
    @in_trap_check
    def mot_population(self):
        """Property that returns the thermal atom count of a job's output

        Returns:
            int: The thermal atom count of a job's output
        """
        return self.thermal_atom_number

    @property
    @in_trap_check
    def thermal_population(self):
        """Property that returns the thermal atom count of a job's output

        Returns:
            int: The thermal atom count of a job's output
        """
        return self.thermal_atom_number

    @property
    @in_trap_check
    def condensed_population(self):
        """Property that returns the condensed atom count of a job's output

        Returns:
            int: The condensed atom count of a job's output
        """
        return self.condensed_atom_number

    @property
    @in_trap_check
    def total_population(self):
        """Property that returns the count of both thermal and condensed atoms of a job's output

        Returns:
            int: The count of both thermal and condensed atoms of a job's output
        """
        return self.thermal_population + self.condensed_population

    @property
    @in_trap_check
    def condensed_fraction(self):
        """Property that returns the condensed fraction of atoms for the condensed population and
        total population of a job's output

        Returns:
            float: The condensed fraction of atoms for the condensed population and the total
                population of a job's output
        """
        return np.round(self.condensed_population / self.total_population, 3)

    @property
    @in_trap_check
    def get_bimodal_fit_parameters(self):
        """Property that returns the TOF fit data for a job's output

        Returns:
            job_schemas.TofFit: The TOF fit data for a job's output
        """
        return self.tof_fit

    def get_image_data(self, image: OutputImageType | None = None) -> np.ndarray | None:
        """Method to retrieve the image data for the specified image type, if no image type is
        provided the job's imaging type will be returned

        Args:
            image (oqtant.schemas.output.OutputImageType | None, optional): The image type to retrieve

        Returns:
            numpy.ndarray: The image data for the specified image type
        """
        in_trap = job_schema.ImageType.IN_TRAP
        tof = job_schema.ImageType.TIME_OF_FLIGHT

        data, shape = None, None
        if image is None:
            # no type specified, assume user wants output image of the correct
            # type based on the image_type used in the job
            image = in_trap if self.image_type == in_trap else tof

        if image == in_trap:
            data = self.it_plot.pixels
            shape = (self.it_plot.rows, self.it_plot.columns)
        elif image == tof:
            data = self.tof_image.pixels
            shape = (self.tof_image.rows, self.tof_image.columns)
        elif image == "MOT":
            data = self.mot_fluorescence_image.pixels
            shape = (
                self.mot_fluorescence_image.rows,
                self.mot_fluorescence_image.columns,
            )
        elif image == "TIME_OF_FLIGHT_FIT":
            if self.image_type == in_trap:
                print("** no fit image available for IN_TRAP **")
                return None
            data = self.tof_fit_image.pixels
            shape = (self.tof_fit_image.rows, self.tof_fit_image.columns)

        output = None
        try:
            output = np.asarray(data).reshape(shape)
        except ValueError:
            print("Could not get image Data.")

        return output

    def get_image_pixcal(self, image: OutputImageType) -> float:
        """Method to get the pixel calibration for the provided image type

        Args:
            image (oqtant.schemas.output.OutputImageType): The image type to retrieve the pixel calibration for

        Returns:
            float: The pixel calibration for the provided image
        """

        def __parse_pixcal(image_name: str) -> float:
            try:
                return getattr(self, image_name).pixcal
            except Exception:
                raise JobReadError(
                    f"{self.job_type} job does not contain a {image} image"
                )

        if image == OutputImageType.TIME_OF_FLIGHT:
            pixcal = __parse_pixcal("tof_image")
        elif image == OutputImageType.IN_TRAP:
            pixcal = __parse_pixcal("it_plot")
        elif image == OutputImageType.MOT:
            pixcal = __parse_pixcal("mot_fluorescence_image")
        elif image == OutputImageType.TIME_OF_FLIGHT_FIT:
            pixcal = __parse_pixcal("tof_fit_image")
        else:
            # raise exception unknown image type
            image_name = (
                "tof_image" if image == OutputImageType.TIME_OF_FLIGHT else "it_plot"
            )
            pixcal = __parse_pixcal(image_name)
        return pixcal

    def get_slice(self, axis: AxisType = "x") -> list[float]:
        """Method that returns a list of data point representing a slice along the specified axis

        Args:
            axis (oqtant.schemas.output.AxisType, optional): The axis along which the take the slice

        Returns:
            list[float]: A list of data points representing the slice along the specified axis
        """
        if axis == "x":
            data = self.tof_x_slice.points
            cut_data = [point["y"] for point in data]
        else:
            data = self.tof_y_slice.points
            cut_data = [point["y"] for point in data]
        return cut_data

    @staticmethod
    def get_image_space(
        datafile: np.ndarray = np.zeros((100, 100)), centered: str = "y"
    ) -> tuple[np.meshgrid, int, int]:
        """Method to generate a numpy.meshgrid of image coordinates

        Args:
            datafile (numpy.ndarray, default): A matrix of optical density data
            centered (str, optional): The orientation of the image

        Returns:
            tuple[numpy.meshgrid, int, int]: The numpy.meshgrid of image coordinates
        """
        lx, ly = np.shape(datafile)
        x, y = np.arange(lx), np.arange(ly)

        if centered == "y":
            x = x - round(lx / 2)
            y = y - round(ly / 2)

        xy_mesh = np.meshgrid(x, y)

        return xy_mesh, lx, ly

    def fit_bimodal_data2D(
        self, xi: list[float] = None, lb: list[float] = None, ub: list[float] = None
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.float64]:
        """Method to perform a fit via a trust region reflective algorithm

        Args:
            xi (list[float] | None, optional): List of fit parameter initial guesses
            lb (list[float] | None, optional): List of fit parameter lower bounds
            ub (list[float] | None, optional): List of fit parameter upper bounds

        Returns:
            tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.float64]: The calculated fit data
        """
        xi = xi if xi else [0.25, 8, 8, 1, 4, 6, 0, 0, 0.02]
        lb = lb if lb else [0, 7, 7, 0, 2, 2, -20, -20, 0]
        ub = ub if ub else [2, 20, 20, 2, 20, 20, 20, 20, 1]

        xy_mesh, _, _ = self.get_image_space()  # TOF_data)

        (X, Y) = xy_mesh
        x = X[0]
        y = Y[:, 0]

        fit_params, cov_mat = opt.curve_fit(
            bimodal_dist_2D, xy_mesh, np.ravel(self.TOF), p0=xi, bounds=(lb, ub)
        )
        fit_residual = self.TOF - bimodal_dist_2D(xy_mesh, *fit_params).reshape(
            np.outer(x, y).shape
        )
        fit_r_squared = 1 - np.var(fit_residual) / np.var(self.TOF)

        return fit_params, cov_mat, fit_residual, fit_r_squared

    def plot_fit_results(
        self,
        fit_params: np.ndarray,
        model: str = "bimodal",
        file_name: str = None,
        plot_title: str = None,
        pix_cal: float = 1.0,
    ) -> None:
        """Method to plot the results of a fit operation

        Args:
            fit_params (numpy.ndarray): List of parameters from a fit operation
            model (str. optional): The shape(?) to use while plotting
            file_name (str | None, optional): The name of the file to write the plot to
            plot_title (str | None, optional): The title of the resulting plot result
            pix_cal (float, optional): The pixel calibration to use while generating the plot
        """

        xy_mesh, lx, ly = self.get_image_space()  # TOF_data)

        (X, Y) = xy_mesh

        if model == "bimodal":
            try:
                m = bimodal_dist_2D(xy_mesh, *fit_params)
            except TypeError as exc:
                raise JobPlotFitMismatchError() from exc
            except Exception as exc:
                raise JobPlotFitError() from exc

        elif model == "gaussian":
            try:
                m = Gaussian_dist_2D(xy_mesh, *fit_params)
            except TypeError as exc:
                raise TypeError(
                    "PLOT FIT RESULTS: mismatched parameters and model type"
                ) from exc
            except Exception as exc:
                raise JobPlotFitError() from exc
        elif model == "TF":
            try:
                m = TF_dist_2D(xy_mesh, *fit_params)
            except TypeError as exc:
                raise JobPlotFitMismatchError() from exc
            except Exception as exc:
                raise JobPlotFitError() from exc
        else:
            print(
                f"PLOT FIT RESULTS: Invalid model specified: {model}.",
                " Select 'bimodal', 'gaussian', or 'TF'",
            )
            return

        m = m.reshape(lx, ly)
        plt.figure()
        plt.imshow(
            m,
            origin="upper",
            cmap="viridis",
            extent=(
                np.min(X) * pix_cal,
                np.max(X) * pix_cal,
                np.min(Y) * pix_cal,
                np.max(Y) * pix_cal,
            ),
        )

        if plot_title is None:
            plot_title = f"job: {self.name}\nTOF fit: {model}"

        plt.title(plot_title)

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    @staticmethod
    def _save_plot_file(plot: plt, file_name: str) -> None:
        """Method to save a plot to a file

        Args:
            plot (matplotlib.pyplot): Module containing the plot to save
            file_name (str): The name of the file to save
        """
        file = f"{file_name}.png"
        try:
            plot.savefig(file)
            print(f"plot saved to file: {file}")
        except (FileNotFoundError, Exception):
            print(f"failed to save plot at {file}")

    def plot_tof(
        self,
        file_name: str = None,
        figsize: tuple[int, int] = (12, 12),
        grid_on: bool = False,
    ) -> None:
        """Method to generate a 2D plot of atom OD

        Args:
            file_name (str | None, optional): The name of the file to write the plot to
            figsize (tuple[int, int], optional): The size of the figure to generate
            grid_on (bool, optional): Flag to show grid lines or not in the plot

        """
        xy_mesh, _, _ = self.get_image_space()  # TOF_data
        (X, Y) = xy_mesh

        fig2D = plt.figure(figsize=figsize)
        ax = fig2D.gca()
        plt2D = plt.imshow(
            self.TOF,
            origin="upper",
            cmap="viridis",
            extent=(
                np.min(X) * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                np.max(X) * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                np.min(Y) * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                np.max(Y) * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
            ),
        )
        plt.grid(b=grid_on)
        plt.colorbar(plt2D, shrink=0.8)

        ax.set_xlabel("x position (microns)", labelpad=15, fontsize=16)
        ax.set_ylabel("y position (microns)", labelpad=15, fontsize=16)
        plt.title("time of flight optical depth", fontsize=16)

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    def plot_slice(
        self, file_name: str = None, axis: AxisType = "x", grid_on: bool = False
    ) -> None:
        """Method to generate a 1D slice plot of atom OD in x or y

        Args:
            file_name (str | None, optional): The name of the file to write the plot to
            axis: (oqtant.schemas.output.AxisType, optional): The axis to use in the plot
            grid_on (bool, optional): Flag to show grid lines or not in the plot
        """
        xy_mesh, lx, ly = self.get_image_space(self.TOF)
        (X, Y) = xy_mesh

        params, *_ = self.fit_bimodal_data2D()
        fitOD = bimodal_dist_2D(xy_mesh, *params)

        Gfit_params = [params[0], params[6], params[7], params[1], params[2], params[8]]
        fitODG = Gaussian_dist_2D(xy_mesh, *Gfit_params)

        # Reshape Fit Distributions to 2D form
        fitOD2D = fitOD.reshape(lx, ly)
        fitODG2D = fitODG.reshape(lx, ly)

        # Define Central slices
        xslice = fitOD2D[int(lx / 2), :]
        yslice = fitOD2D[:, int(ly / 2)]
        xsliceG = fitODG2D[int(lx / 2), :]
        ysliceG = fitODG2D[:, int(ly / 2)]

        if axis == "x":
            xsliceD = self.TOF[int(len(X[1]) / 2), :]
            xslice = fitOD2D[int(len(X[1]) / 2), :]
            xsliceG = fitODG2D[int(len(X[1]) / 2), :]
            plt.title("x slice", fontsize=16)
            plt.plot(
                X[1] * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                xsliceD,
                "ok",
            )
            plt.plot(
                X[1] * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                xslice,
                "b",
            )
            plt.plot(
                X[1] * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                xsliceG,
                "r",
            )
        elif axis == "y":
            ysliceD = self.TOF[:, int(len(Y[1]) / 2)]
            yslice = fitOD2D[:, int(len(Y[1]) / 2)]
            ysliceG = fitODG2D[:, int(len(Y[1]) / 2)]
            plt.title("y slice", fontsize=16)
            plt.plot(
                Y[:, 1] * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                ysliceD,
                "ok",
            )
            plt.plot(
                Y[:, 1] * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                yslice,
                "b",
            )
            plt.plot(
                Y[:, 1] * self.get_image_pixcal(OutputImageType.TIME_OF_FLIGHT),
                ysliceG,
                "r",
            )
        else:
            raise ValueError("input either x or y")

        plt.grid(b=grid_on)
        plt.xlabel("x position (microns)", labelpad=15, fontsize=16)
        plt.ylabel("optical depth", labelpad=15, fontsize=16)

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    def plot_it(
        self,
        file_name: str = None,
        figsize: tuple[int, int] = (12, 12),
        grid_on=False,
    ) -> None:
        """Method to plot an in-trap image output

        Args:
            file_name (str | None, optional): The name of the file to write the plot to
            figsize (tuple[int, int], optional): The size of the figure to generate
            grid_on (bool): Show grid in plot?
        """
        it_od = self.get_image_data("IN_TRAP")

        pixcal = self.get_image_pixcal(image="IN_TRAP")

        plt.figure(figsize=figsize)
        plt.title("in-trap optical depth")
        it_plot = plt.imshow(
            it_od,
            origin="upper",
            cmap="viridis",
            extent=[-256 * pixcal, 256 * pixcal, -74 * pixcal, 74 * pixcal],
        )
        plt.xlabel("x position (microns)")
        plt.ylabel("y position (microns)")
        plt.grid(visible=grid_on)
        plt.colorbar(it_plot, shrink=0.25)
        plt.show()

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    # This function plots the optical depth as a 3D surface with projected density contours
    def plot_tof_3d(
        self,
        file_name: str = None,
        view_angle: int = -45,
        figsize: tuple[int, int] = (10, 10),
    ) -> None:
        """Method to generate a 3D slice plot of atom OD

        Args:
            file_name (str | None, optional): The name of the file to write the plot to
            view_angle (int, optional): Azimuthal/horizontal angle of "camera" view
            figsize (tuple[int, int], optional): The size of the figure to generate
        """

        fig3d = plt.figure(figsize=figsize)
        ax = fig3d.gca(projection="3d")

        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter("%.02f"))
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False

        # Set axis labels
        ax.set_xlabel("x position (microns)", labelpad=10)
        ax.set_ylabel("y position (microns)", labelpad=10)
        ax.set_zlabel("optical depth", labelpad=10)

        # rotate the axes and update
        ax.view_init(30, view_angle)

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()


class OqtantPlotOutput(OqtantOutput, job_schema.PlotOutput):
    ...


class OqtantNonPlotOutput(OqtantOutput, job_schema.NonPlotOutput):
    ...


def round_sig(x: float, sig: int = 2) -> float:
    """Method to round a number to a specified number of significant digits

    Args:
        x (float): The number to be rounded
        sig (int, optional): The number of significant digits

    Returns:
        float: The rounded number
    """
    return round(x, sig - int(floor(log10(abs(x)))) - 1)


def TF_dist_2D(
    xy_mesh: tuple[np.ndarray, np.ndarray],
    TFpOD: float,
    xc: float,
    yc: float,
    rx: float,
    ry: float,
    os: float,
) -> np.ndarray:
    """Method to sample a 2D Thomas-Fermi distribution with given parameters on a grid of coordinates

    Args:
        xy_mesh (tuple[numpy.ndarray, numpy.ndarray]): Matrix containing meshgrid of image coordinates
        TFpOD (float): Thomas-Fermi peak optical density
        xc (float): Cloud center along the x direction (along gravity)
        yc (float): Cloud center along the y direction
        rx (float): Thomas-Fermi radius along the x direction
        ry (float): Thomas-Fermi radius along the y direction (along gravity)
        os (float): Constant offset

    Returns:
        numpy.ndarray: a 2D array of samples from a Thomas Fermi distribution
    """

    # unpack 1D list into 2D x and y coords
    (x, y) = xy_mesh

    # Simplify Thomas-Fermi expression
    A = 1 - ((y - yc) / ry) ** 2 - ((x - xc) / rx) ** 2

    # make 2D Thomas-Fermi distribution
    OD = np.real(TFpOD * np.maximum(np.sign(A) * (np.abs(A)) ** (3 / 2), 0)) + os

    # flatten the 2D Gaussian down to 1D
    return OD.ravel()


def Gaussian_dist_2D(
    xy_mesh: tuple[np.ndarray, np.ndarray],
    GpOD: float,
    xc: float,
    yc: float,
    sigx: float,
    sigy: float,
    os: float,
) -> np.ndarray:
    """Method to sample a 2D Gaussian distribution with given parameters on a grid of coordinates

    Args:
        xy_mesh (tuple[numpy.ndarray, numpy.ndarray]): Matrix containing meshgrid of image coordinates
        GpOD (float): Gaussian peak optical density
        xc (float): Cloud center along the x direction (along gravity)
        yc (float): Could center along the y direction
        sigx (float): Gaussian spread along the x direction
        sigy (float): Gaussian spread along the y direction (along gravity)
        os (float): Constant offset

    Returns:
        numpy.ndarray: a 2D array of samples from a Gaussian distribution
    """

    (x, y) = xy_mesh

    OD = (
        GpOD * np.exp(-0.5 * ((y - yc) / sigy) ** 2 - 0.5 * ((x - xc) / sigx) ** 2) + os
    )
    return OD.ravel()


def bimodal_dist_2D(
    xy_mesh: tuple[np.ndarray, np.ndarray],
    GpOD: float,
    sigx: float,
    sigy: float,
    TFpOD: float,
    rx: float,
    ry: float,
    xc: float,
    yc: float,
    os: float,
):
    """Method to sample a bimodal (Thomas-Fermi + Gaussian) distribution with given parameters on a grid of coordinates

    Args:
        xy_mesh (tuple[numpy.ndarray, numpy.ndarray]): Matrix containing meshgrid of image coordinates
        GpOD (float): Gaussian peak optical density
        sigx (float): Gaussian spread along the x direction
        sigy (float): Gaussian spread along the y direction (along gravity)
        TFpOD (float): Thomas-Fermi peak optical density
        rx (float): Thomas-Fermi radius along the x direction
        ry (float): Thomas-Fermi radius along the y direction (along gravity)
        xc (float): Cloud center along the x direction (along gravity)
        yc (float): Cloud center along the y direction
        os (float): Constant offset

    Returns:
        numpy.ndarray: a 2D array of samples from a bimodal (Thomas-Fermi + Gaussian) distribution
    """

    return Gaussian_dist_2D(xy_mesh, GpOD, xc, yc, sigx, sigy, os) + TF_dist_2D(
        xy_mesh, TFpOD, xc, yc, rx, ry, os
    )



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import warnings
from typing import Annotated

from bert_schemas import job as job_schema
from bert_schemas.job import RfInterpolationType, interpolate_1d_list
from pydantic import BaseModel, ConfigDict, Field, model_validator


class ConversionError(Exception): ...


class RfSequence(job_schema.RfEvaporation):
    """A class that represents a sequence of radio frequency powers/frequencies in time"""

    @classmethod
    def new(
        cls,
        times: list[float] = [0],
        powers: list[float] = [0],
        frequencies: list[float] = [0],
        interpolation: job_schema.InterpolationType = "LINEAR",
    ) -> RfSequence:
        """Method to create a new RfSequence object

        Args:
            times (list[float], optional): List of times, in ms
            powers (list[float], optional): List of powers, in MHz
            frequencies (list[float], optional): List of powers, mW
            interpolation (bert_schemas.job.InterpolationType, optional): Interpolation type of the RF sequence

        Returns:
            RfSequence: A new RfSequence object
        """
        return cls(
            times_ms=times,
            powers_mw=powers,
            frequencies_mhz=frequencies,
            interpolation=interpolation,
        )

    @classmethod
    def from_input(cls, rf_evaporation: job_schema.RfEvaporation) -> RfSequence:
        """Method to create a RfSequence object using the input values of a job

        Args:
            rf_evaporation (bert_schemas.job.RfEvaporation): The input values

        Returns:
            RfSequence: A new RfSequence object created using the input data
        """
        return cls(**rf_evaporation.model_dump())

    def get_frequencies(self, times: list[float]) -> list[float]:
        """Method to calculate the RF evaporation frequencies of a RfSequence object at the specified times

        Args:
            times (list[float]): The times, in ms, at which the RF frequencies are calculated

        Returns:
            list[float]: The calculated frequencies, in MHz, at the specified times
        """
        return interpolate_1d_list(
            self.times_ms,
            self.frequencies_mhz,
            times,
            self.interpolation,
        )

    def get_powers(self, times: list[float]) -> list[float]:
        """Method to calculate the RF evaporation powers at of a RfSequence object at the specified times

        Args:
            times (list[float]): The times, in ms, at which the RF powers are calculated

        Returns:
            list[float]: The RF powers, in mW, at the specified times
        """
        return interpolate_1d_list(self.times_ms, self.powers_mw, times, self.interpolation)


class RfEvap(RfSequence):
    """A class that represents the forced RF evaporation that cools atoms to quantum degeneracy."""

    @classmethod
    def new(
        cls,
        times: list[float] = [0],
        powers: list[float] = [0],
        frequencies: list[float] = [0],
        interpolation: job_schema.InterpolationType = "LINEAR",
    ) -> RfEvap:
        """Method to create a new RfEvap object

        Args:
            times (list[float], optional): List of times, in ms
            powers (list[float], optional): List of powers, in MHz
            frequencies (list[float], optional): List of powers, mW
            interpolation (bert_schemas.job.InterpolationType, optional): Interpolation type of the RF evaporation

        Returns:
            RfEvap: A new RfEvap object
        """
        return cls(
            times_ms=[t - max(times) for t in times],
            powers_mw=powers,
            frequencies_mhz=frequencies,
            interpolation=interpolation,
        )

    @classmethod
    def from_input(cls, rf_evaporation: job_schema.RfEvaporation) -> RfEvap:
        """Method to create a RfEvap object using the input values of a job

        Args:
            rf_evaporation (bert_schemas.job.RfEvaporation): The input values

        Returns:
            RfEvap: A new RfEvap object created using the input data
        """
        rf_evap_times = [t for t in rf_evaporation.times_ms if t <= 0.0]
        if rf_evap_times == []:
            raise ConversionError()

        rf_evap_freqs = [f for t, f in zip(rf_evaporation.times_ms, rf_evaporation.frequencies_mhz) if t <= 0.0]

        rf_evap_pows = [p for t, p in zip(rf_evaporation.times_ms, rf_evaporation.powers_mw) if t <= 0.0]
        return cls.new(
            times=rf_evap_times,
            frequencies=rf_evap_freqs,
            powers=rf_evap_pows,
            interpolation=rf_evaporation.interpolation,
        )


# A RfShield is a CONSTANT evaporation occurring during the entire experimental phase
# any non-negative time in the rf_evaporation object of a program indicates a
# rf shield is desired for the entire duration of the experiment stage
class RfShield(BaseModel):
    """A class that represents an RF shield (at fixed frequency and power)
    being applied during the 'experiment' phase/stage."""

    times_ms: Annotated[
        list[Annotated[int, Field(ge=-2000, le=200)]],
        Field(min_length=1, max_length=20),
    ] = list(range(-1600, 400, 400))
    frequencies_mhz: Annotated[
        list[Annotated[float, Field(ge=0.0, le=25.0)]],
        Field(min_length=1, max_length=20),
    ]
    powers_mw: Annotated[
        list[Annotated[float, Field(ge=0.0, le=1000.0)]],
        Field(min_length=1, max_length=20),
    ]
    interpolation: RfInterpolationType
    model_config = ConfigDict(validate_assignment=True, extra="forbid")

    @model_validator(mode="after")
    def cross_validate(self) -> RfShield:
        if not len(self.times_ms) == len(self.frequencies_mhz) == len(self.powers_mw):
            raise ValueError("RfEvaporation data lists must have the same length.")

        if self.times_ms != sorted(self.times_ms):
            warnings.warn(
                "Evaporation times_ms list must be naturally ordered, re-ordering.",
                stacklevel=2,
            )
            self.times_ms, self.frequencies_mhz, self.powers_mw = zip(
                *sorted(
                    zip(
                        self.times_ms,
                        self.frequencies_mhz,
                        self.powers_mw,
                    )
                )
            )
        return self

    @classmethod
    def new(
        cls,
        lifetime: float,
        frequency: float,
        power: float,
        interpolation: job_schema.InterpolationType = "LINEAR",
    ) -> RfShield:
        """Method to create a new RfShield object

        Args:
            lifetime (float): Lifetime of the shield, in ms
            frequency (float | None): Frequency of the shield, in MHz
            power (float | None): Power of the shield, in mW
            interpolation (bert_schemas.job.InterpolationType, optional): Interpolation type of the shield

        Returns:
            RfShield: A new RfShield object
        """
        return cls(
            times_ms=[lifetime],
            powers_mw=[power],
            frequencies_mhz=[frequency],
            interpolation=interpolation,
        )

    @classmethod
    def from_input(cls, rf_evaporation: job_schema.RfEvaporation) -> RfShield:
        """Method to create a RfShield object using the input values of a job

        Args:
            rf_evaporation (bert_schemas.job.RfEvaporation): The input values

        Returns:
            RfShield: A new RfShield object created using the input data
        """
        if rf_evaporation.times_ms[-1] <= 0:
            raise ConversionError()
        else:
            return cls.new(
                lifetime=rf_evaporation.times_ms[-1],
                frequency=rf_evaporation.frequencies_mhz[-1],
                power=rf_evaporation.powers_mw[-1],
                interpolation=rf_evaporation.interpolation,
            )

    @property
    def lifetime(self) -> float:
        """Property to get the lifetime value of a RfShield object

        Returns:
            float: The amount of time, in ms, that the shield will exist
        """
        return self.times_ms[0]

    @property
    def frequency(self) -> float:
        """Property to get the frequency value of a RfShield object

        Returns:
            float: The shield's frequency, in MHz
        """
        return self.frequencies_mhz[0]

    def frequencies(self, times: list[float]) -> list[float]:
        """Method to generate a list of frequencies using the provided list of times

        Args:
            times (list[float]): The times, in ms, at which the frequencies are generated

        Returns:
            list(float): The calculated frequencies, in MHz, at the specified times
        """
        return [self.frequencies_mhz[0]] * len(times)

    @property
    def power(self) -> float:
        """Property to get the power value of a RfShield object

        Returns:
            float: The shield's power, in mW
        """
        return self.powers_mw[0]

    def powers(self, times: list[float]) -> list[float]:
        """Method to generate a list of powers using the provided list of times

        Args:
            times (list[float]): The times, in ms, at which the powers are generated

        Returns:
            list(float): The calculated powers, in mW, at the specified times
        """
        return [self.powers_mw[0]] * len(times)



# Copyright 2024 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import numpy as np


@dataclass
class ThreeDimGrid:
    """
    'ThreeDimGrid' Defines a two dimensional grid space in cylindrical coordinates with axial symmetry.
    Nx and Nr are the number of points in the x and r directions
    Lx and Lr are the lenghts of the x and z dimensions
    tof_nf toggles the near field grids
    tof_ff toggles the far field grids
    """

    Nx: int
    Nr: int
    Lx: int
    Lr: int
    dr: int
    dz: int
    dy: int

    def __init__(self, tof_nf: bool = False, tof_ff: bool = False):
        # parameters for 3D space
        if tof_nf and tof_ff:
            raise ValueError("tof_nf and tof_ff cannot both be True")
        if tof_nf:
            self.Nx = 500
            self.Nr = 300
            self.Lx = 50
            self.Lr = 30
        elif tof_ff:
            self.Nx = 500
            self.Nr = 500
            self.Lx = 70
            self.Lr = 50
        else:
            self.Nx = 400
            self.Nr = 40
            self.Lx = 40
            self.Lr = 4

        self.dx = self.Lx / self.Nx
        self.dr = self.Lr / self.Nr

        # defining cylindrical grids using np.meshgrid
        self.x_1d, self.r_1d = np.linspace(
            -self.Lx / 2, self.Lx / 2, self.Nx
        ), np.linspace(self.dr / 2, self.Lr - self.dr / 2, self.Nr)

        self.x, self.r = np.meshgrid(self.x_1d, self.r_1d)

        # defining the equivalent cartesian grids for constructing the column densities
        self.z, self.y = np.meshgrid(
            np.linspace(-self.Lr + self.dr / 2, self.Lr - self.dr / 2, 2 * self.Nr),
            np.linspace(-self.Lr + self.dr / 2, self.Lr - self.dr / 2, 2 * self.Nr),
        )
        self.dz, self.dy = self.dr, self.dr
